<?php
    /**
     * @author Dima_tomsk
     * @copyright Seiros LLC, 2006
     * @version 0.1
     */

    /**
     * TeMPLate error
     */
    define( 'TMPL_NO_ERROR'          , 0 );
    define( 'TMPL_UNKNOWN_ERROR'     , 1 );
    define( 'TMPL_DIR_NOT_EXISTS'    , 2 );
    define( 'TMPL_FILE_NOT_EXISTS'   , 3 );
    define( 'TMPL_WRONG_PARAMS'      , 4 );
    define( 'TMPL_TAG_PARSE_ERROR'   , 5 );
    define( 'TMPL_TAG_REQUIRE_PARAM' , 6 );
    define( 'TMPL_TAG_NOT_CLOSED'    , 7 );
    define( 'TMPL_TAG_LEVEL_ERROR'   , 8 );
    define( 'TMPL_NO_CONTENT'        , 9 );

    /**
     * TeMPLate ERRor type
     */
    define( 'TMPLERR_NOTICE' , 1 );
    define( 'TMPLERR_FATAL'  , 2 );

    /**
     * ESCape MethoD
     */
    define( 'ESCMD_HTML', 'HTML' );
    define( 'ESCMD_NONE', 'NONE' );
    define( 'DEFAULT_ESCAPE_METHOD', ESCMD_HTML );

    /**
     * Дерево возможных вызовов функций ( ... - рекурсия )
     *
     *  CTemplate()
     *      setDir()
     *      parse()
     *
     *  error()
     *      errorString()
     *
     *  errorString()
     *
     *  output()
     *      _processTags()
     *          $_functions[  xxx ]
     *              _processTags()...
     *
     *  param()
     *
     *  parse()
     *      _parseLevel()
     *          _parseString()
     *              _nextParam()
     *          _parseLevel()...
     *
     *  setDir()
     *
     *  value()
     *
     */
    class CTemplate {

        /**
         * @var string
         * @access private
         */
        var $_dirname = '';

        /**
         * открывающий тег ( например, может быть "<!-- TMPL_" или "{" )
         *
         * @var string
         * @access private
         */
        var $_leftlim  = '<TMPL_';

        /**
         * Длина строки $_leftlim
         *
         * @var integer
         * @access private
         */
        var $_ll;

        /**
         * закрывающий тег ( например, может быть "<!-- /TMPL_" или "{/" )
         *
         * @var string
         * @access private
         */
        var $_cleftlim = '</TMPL_';

        /**
         * Длина строки $_cleftlim
         *
         * @var integer
         * @access private
         */
        var $_cl;

        /**
         * строка, закрывающая тег ( например, может быть "-->" или "}" )
         *
         * @var string
         * @access private
         */
        var $_rightlim = '>';

        /**
         * Длина строки $_rightlim
         *
         * @var integer
         * @access private
         */
        var $_rl;

        /**
         * Код ошибки, устанавливается в случае нефатальной ошибки
         *
         * @var integer
         * @access public
         */
        var $errno = 0;

        /**
         * Строка с описанием последней ошибки
         *
         * @var string
         * @access public
         */
        var $lasterror;

        /**
         * массив переменых для задания параметров шаблона
         *
         * @var array
         * @access private
         */
        var $_values = array();

        /**
         * Массив "строк" шаблона
         *
         * @var array
         * @access private
         */
        var $_lines = array();

        /**
         * Строка до первого тега в шаблоне
         *
         * @var string
         * @access private
         */
        var $_prefix = '';

        /**
         * Строка, в которую собирается результат подстановки в шаблон переменных
         *
         * @var string
         * @access private
         */
        var $_output = '';

        /**
         * Массив переводов ошибок
         *
         * @var array
         * @access private
         */
        var $_errors = array(
                TMPL_NO_ERROR          => 'Нет ошибки',
                TMPL_UNKNOWN_ERROR     => 'Неизветная ошибка',
                TMPL_DIR_NOT_EXISTS    => 'Директория не существует',
                TMPL_FILE_NOT_EXISTS   => 'Файл не существует',
                TMPL_WRONG_PARAMS      => 'Параметр должен быть либо массив, либо null',
                TMPL_TAG_PARSE_ERROR   => 'Ошибка анализа тега',
                TMPL_TAG_REQUIRE_PARAM => 'В теге отсутствует необходимый параметр',
                TMPL_TAG_NOT_CLOSED    => 'Тег не закрыт',
                TMPL_TAG_LEVEL_ERROR   => 'Ошибка в структуре тегов',
            );

        /**
         * массив допустримых в идентификаторе символов
         * 0 означает, что идентификатор может начинаться с этого символа
         * 1 - нет.
         * Учитывая, что на индексы ассоциативных массивов не наклыдывается
         * такого ограничения, это разделение может оказаться неуместным, в этом
         * случае вы вольны поступить со значениями элементов так, как хотите.
         *
         * @var array
         * @access private
         */
        var $_chars = array( 'a' => 0, 'b' => 0, 'c' => 0, 'd' => 0, 'e' => 0, 'f' => 0, 'g' => 0, 'h' => 0, 'i' => 0, 'j' => 0, 'k' => 0, 'l' => 0, 'm' => 0, 'n' => 0, 'o' => 0, 'p' => 0, 'q' => 0, 'r' => 0, 's' => 0, 't' => 0, 'u' => 0, 'v' => 0, 'w' => 0, 'x' => 0, 'y' => 0, 'z' => 0,
                             'A' => 0, 'B' => 0, 'C' => 0, 'D' => 0, 'E' => 0, 'F' => 0, 'G' => 0, 'H' => 0, 'I' => 0, 'J' => 0, 'K' => 0, 'L' => 0, 'M' => 0, 'N' => 0, 'O' => 0, 'P' => 0, 'Q' => 0, 'R' => 0, 'S' => 0, 'T' => 0, 'U' => 0, 'V' => 0, 'W' => 0, 'X' => 0, 'Y' => 0, 'Z' => 0,
                             '_' => 0,
                             '0' => 1, '1' => 1, '2' => 1, '3' => 1, '4' => 1, '5' => 1, '6' => 1, '7' => 1, '8' => 1, '9' => 1 );

        /**
         * Массив поддерживаемых тегов и их обязательные параметры
         *
         * @var array
         * @access private
         */
        var $_tags = array( 'VAR'     => array( 'NAME' => 1 ),
                            'LOOP'    => array( 'NAME' => 1 ),
                            'FOREACH' => array( 'NAME' => 1 ),
                            'IF'      => array( 'NAME' => 1 ),
                            'UNLESS'  => array( 'NAME' => 1 ),
                            'INCLUDE' => array( 'NAME' => 1 ),
                            'ELSE'    => array() );

        /**
         * Массив имен функций обрабатывающих теги
         *
         * @var array
         * @access private
         */
        var $_functions = array( 'VAR'     => '_processVar',
                                 'LOOP'    => '_processLoop',
                                 'FOREACH' => '_processLoop',
                                 'IF'      => '_processIf',
                                 'UNLESS'  => '_processUnless',
                                 'INCLUDE' => '_processInclude' );

        /**
         * Массив тегов разбивающих шаблон на блоки
         *
         * @var array
         * @access private
         */
        var $_blocktags = array( 'LOOP'    => 1,
                                 'FOREACH' => 1,
                                 'IF'      => 1,
                                 'ELSE'    => 1,
                                 'UNLESS'  => 1 );

        // {{{ CTemplate()
        /**
         * Конструктор
         *
         * @param string $dirname
         * @param string $filename
         * @return CTemplate
         */
        function CTemplate( $dirname = null, $filename = null )
        {
            $this->_ll = strlen( $this->_leftlim );
            $this->_cl = strlen( $this->_cleftlim );
            $this->_rl = strlen( $this->_rightlim );
            if ( isset( $dirname ) ) {
                $this->setDir( $dirname );
            }
            if ( isset( $filename ) ) {
                $this->parse( array( 'filename' => $filename ) );
            }
        }
        // }}} CTemplate()

        // {{{ errorString()
        /**
         * Функция возвращает строку с описанием ошибки по коду ошибки
         *
         * @param integer $code
         * @return string
         */
        function errorString( $code )
        {
            if ( isset( $this->_errors[ $code ] ) ) {
                return $this->_errors[ $code ];
            }
            else {
                return $this->_errors[ TMPL_UNKNOWN_ERROR ] . ' ' . $code;
            }
        }
        // }}} errorString()

        // {{{ error()
        /**
         * Функция обработки ошибки
         *
         * @param integer $code
         * @param string $line
         */
        function error( $code = -1, $line = '', $errlevel = TMPLERR_FATAL )
        {
            $this->lasterror = $this->errorString( $code ) . ', ' . $line;
            if( $errlevel === TMPLERR_FATAL ) {
                die( $this->lasterror );
            }
            else {
                $this->errno = $code;
                return false;
            }
        }
        // }}} error()

        // {{{ setDir()
        /**
         * Установка директории
         *
         * @param string $dirname
         * @return boolean
         */
        function setDir( $dirname )
        {
            if ( is_dir( $dirname ) ) {
                $end = substr( $dirname, -1 );
                $this->_dirname = $dirname . ( ( $end == '/' || $end == '\\' ) ? '' : '/'  );
                return true;
            }
            else {
                return $this->error( TMPL_DIR_NOT_EXISTS, $dirname );
            }
        }
        // }}} setDir()

        // {{{ _min()
        /**
         * Параметры и результат, integer либо boolean
         *
         * @param mixed $v1
         * @param mixed $v2
         * @return mixed
         */
        function _min( $v1, $v2 )
        {
            if ( $v1 === false ) {
                return $v2;
            }
            if ( $v2 === false ) {
                return $v1;
            }
            return ( $v1 < $v2 ? $v1 : $v2 );
        }
        // }}} _min()

        // {{{ _parseLevel()
        /**
         * Рекурсивный анализ шаблона
         *
         * @param string $text
         * @param array $level
         * @param integer $o
         * @param integer $c
         * @return mixed
         */
        function _parseLevel( & $text, & $level, & $o, & $c )
        {
            $level[ 'underlying' ] = null;
            while ( $o !== false || $c !== false ) {
                $block = null;
                if ( $c === false || ( $o !== false && $o < $c ) ) {
                    $block[ 'open' ] = 1;
                    $n = $o + $this->_ll;
                    $o = strpos( $text, $this->_leftlim, $n );
                }
                else {
                    $block[ 'open' ] = 0;
                    $n = $c + $this->_cl;
                    $c = strpos( $text, $this->_cleftlim, $n );
                }
                $m = $this->_min( $c, $o );
                $e = strpos( $text, $this->_rightlim, $n );
                if ( !$e || ( $m && $e > $m ) ) {
                    return $this->error( TMPL_TAG_NOT_CLOSED, substr( $text, $n, 100 ) );
                }
                else {
                    $block[ 'tail' ] = substr( $text, $n, $e + $this->_rl - $n );
                    $this->_parseString( $block );
                    $n = $e + $this->_rl;
                }
                $block[ 'tail' ] = ( !$m ? substr( $text, $n ) : substr( $text, $n, $m - $n ) );

                if ( isset( $this->_blocktags[ $block[ 'tag' ] ] ) ) {
                    if ( $block[ 'open' ] == 0 || $block[ 'tag' ] == 'ELSE' ) {
                        return $block;
                    }
                    else {
                        $tmp = $this->_parseLevel( $text, $block, $o, $c );
                        if( $this->errno ) {
                            return false;
                        }
                        if ( $tmp[ 'tag' ] == 'ELSE' ) {
                            $block[ 'elsetail' ] = $tmp[ 'tail' ];
                            $elseblock = null;
                            $tmp = $this->_parseLevel( $text, $elseblock, $o, $c );
                            if( $this->errno ) {
                                return false;
                            }
                            $block[ 'else' ] = & $elseblock[ 'underlying' ];
                            unset( $elseblock );
                        }
                        if ( $tmp[ 'tag' ] == $block[ 'tag' ] && $tmp[ 'open' ] == 0 ) {
                            $block[ 'endtail' ] = $tmp[ 'tail' ];
                        }
                        else {
                            return $this->error( TMPL_TAG_LEVEL_ERROR . __FILE__ );
                        }
                    } // else
                } // if ( isset( $this->_blocktags[ $block[ 'tag' ] ] ) )
                $block[ 'parent' ] = & $level;
                $level[ 'underlying' ][] = & $block;
                unset( $block );
            } // while ( $o !== false || $c !== false )
            return null;
        }
        // }}} _parseLevel()

        // {{{ parse()
        /**
         * @param array $param
         * @return boolean
         */
        function parse( $param )
    		{
      			if( ! is_array($param) )
      			{
      				$this->parse( array('filename' => $param) );
      			}
            $this->errno = 0;
            $text = null;
            if( isset( $param[ 'filename' ] ) ) {
                if( is_file( $this->_dirname . $param[ 'filename' ] ) ) {
                    $text = implode( '', file( $this->_dirname . $param[ 'filename' ] ) );
                }
                else {
                    return $this->error( TMPL_FILE_NOT_EXISTS, $this->_dirname . $param[ 'filename' ] );
                }
            } elseif( isset( $param[ 'content' ] ) ) {
                $text = & $param[ 'content' ];
            }
            if ( $text !== null ) {
                $o = strpos( $text, $this->_leftlim );
                $c = strpos( $text, $this->_cleftlim );
                $m = $this->_min( $o, $c );
                $this->_lines = array();
                if ( $m !== false ) {
                    $this->_prefix = substr( $text, 0, $m );
                    if ( $this->_parseLevel( $text, $this->_lines, $o, $c ) != null ) {
                        return $this->error( TMPL_TAG_LEVEL_ERROR,
                                ( $o === false ? 'end' : substr( $text, $o, 100 ) ) . ' : ' .
                                ( $c === false ? 'end' : substr( $text, $c, 100 ) ) );
                    }
                }
                else {
                    $this->_prefix = &$text;
                }
                unset( $text );
                return true;
            } // if ( $text !== null )
            else {
                return $this->error( TMPL_NO_CONTENT );
            }
        }
        // }}} parse()

        // {{{ param()
        /**
         * @param array $array
         * @return boolean
         */
        function param( $array )
        {
            if ( is_array( $array ) ) {
                reset( $array );
                while ( list( $key, $value ) = each( $array ) ) {
                	$this->_values[ $key ] = $value;
                }
            }
            elseif ( isset( $array ) ) {
                return $this->error( TMPL_WRONG_PARAMS, $array );
            }
            return true;
        }
        // }}} param()

        // {{{ value()
        /**
         * @param mixed $key
         * @param mixed $value
         */
        function value( $key, $value )
        {
            $this->_values[ $key ] = $value;
        }
        // }}} value()

        // {{{ output()
        /**
         * @return string
         */
        function output()
        {
            return $this->_prefix . $this->_processTags( $this->_lines[ 'underlying' ] );
        }
        // }}} output()

        // {{{ _nextParam()
        /**
         * @param stirng $str
         * @return mixed
         */
        function _nextParam( & $str )
        {
            $i = -1;
            $l = strlen( $str );
            $name = null;
            $type = null;
            while ( ++$i < $l && $str{ $i } == ' ' );
            if ( $i < $l ) {
                $ch = $str{ $i };
                if ( $ch == '=' ) {
                    $type = '=';
                }
                elseif ( isset( $this->_chars[ $ch ] ) && 0 == $this->_chars[ $ch ] ) {
                    $j = $i;
                    while ( ++$i < $l && isset( $this->_chars[ $str{ $i } ] ) );
                    $name = substr( $str, $j, $i - $j );
                    $type = 'name';
                    $i--;
                }
                elseif ( $ch == '\'' || $ch == '"' ) {
                    if ( ++$i < $l ) {
                        $j = $i;
                        // modify 23.06 последовательность символов в кавычках не зависит от $_chars
                        while ( ++$i < $l && $ch != $str{ $i } );
                        $name = substr( $str, $j, $i - $j );
                        if ( $i >= $l || $str{ $i } != $ch ) {
                            return $this->error( TMPL_TAG_PARSE_ERROR, __LINE__ . $str );
                        }
                        $type = 'name';
                    }
                    else {
                        return $this->error( TMPL_TAG_PARSE_ERROR, __LINE__ . $str );
                    }
                } // elseif ( $ch == '\'' || $ch == '"' )
                else {
                    $type = $this->_rightlim;
                }
            } // if ( $i < $l )
            if ( isset( $type ) ) {
                $str = substr( $str, ++$i );
            }
            else {
                return $this->error( TMPL_TAG_PARSE_ERROR, __LINE__ . $str );
            }
            return array( 'type' => $type, 'name' => $name );
        }
        // }}} _nextParam()

        // {{{ _parseString()
        /**
         * @param array $block
         * @return boolean
         */
        function _parseString( & $block )
        {
            if( ( $param = $this->_nextParam( $block[ 'tail' ] ) ) === false ) {
                return false;
            }
            if ( isset( $this->_tags[ $param[ 'name' ] ] ) ) {
                $tag = $block[ 'tag' ] = $param[ 'name' ];
                $state = 0;
                $tmpvalue = null;
                $tmpkey   = null;
                do { // while ( $t != null )
                    if ( $state != 3 ) {
                        $param = $this->_nextParam( $block[ 'tail' ] );
                    }
                    else {
                        if ( isset( $tmpvalue ) ) {
                            $block[ 'NAME' ] = $tmpvalue;
                        }
                        if ( $block[ 'open' ] ) {
                            reset( $this->_tags[ $tag ] );
                            while ( list( $k ) = each( $this->_tags[ $tag ] ) ) {
                                if ( !isset( $block[ $k ] ) ) {
                                    return $this->error( TMPL_TAG_REQUIRE_PARAM, $k . ' ' . $block[ 'tail' ] );
                                }
                            }
                        }
                        return true;
                    } // else
                    $t = $param[ 'type' ];
                    switch ( $state ) {
                        case 0:
                            if ( $t === 'name' ) {
                                $tmpvalue = $param[ 'name' ];
                                $state = 1;
                            }
                            elseif ( $t === $this->_rightlim ) {
                                $state = 3;
                            }
                            else {
                                $t = null;
                            }
                            break;
                        case 1:
                            if ( $t === 'name' ) {
                                $block[ 'NAME' ] = $tmpvalue;
                                $tmpvalue = $param[ 'name' ];
                            }
                            elseif ( $t === '=' ) {
                                $tmpkey = $tmpvalue;
                                $state = 2;
                            }
                            elseif ( $t === $this->_rightlim ) {
                                $state = 3;
                            }
                            else {
                                $t = null;
                            }
                            break;
                        case 2:
                            if ( $t === 'name' ) {
                                $block[ $tmpkey ] = $param[ 'name' ];
                                unset( $tmpvalue );
                                $state = 0;
                            }
                            else {
                                $t = null;
                            }
                            break;
                    } // switch ( $state )
                } while ( $t != null );
            } // if ( isset( $this->_tags[ $param[ 'name' ] ] ) )
            return $this->error( TMPL_TAG_PARSE_ERROR, __LINE__ . $block[ 'tail' ] );
        }
        // }}} _parseString()

        // {{{ _processTags()
        /**
         * @param array $tags
         * @return string
         */
        function _processTags( & $tags )
        {
            $res = '';
            if( is_array( $tags ) ) {
                reset( $tags );
                while( list( $key ) = each( $tags ) ) {
                    $f = $this->_functions[ $tags[ $key ][ 'tag' ] ];
                    $res .= $this->$f( $tags[ $key ] );
                }
            }
            return $res;
        }
        // }}} _processTags()

        // {{{ _getVar()
        /**
         * @param array $block
         * @param string $name
         * @return mixed
         */
        function &_getVar( & $block, $name )
        {
            $b = & $block;
            do {
                if( isset( $b[ 'values' ][ $name ] ) ) {
                    return $b[ 'values' ][ $name ];
                }
                if( isset( $b[ 'parent' ] ) ) {
                    $b = & $b[ 'parent' ];
                }
                else {
                    break;
                }
            } while( 1 );

            if( isset( $this->_values[ $name ] ) ) {
                return $this->_values[ $name ];
            }

            $res = null;
            return $res;
        }
        // }}} _getVar()

        // {{{ _processVar()
        /**
         * @param array $block
         * @return string
         */
        function _processVar( & $block )
        {
            $res = $this->_getVar( $block, $block[ 'NAME' ] );
            if( is_object( $res ) && is_callable( array( $res, 'output' ) ) ) {
                $res = $res->output();
            }
            else {
                if ( !isset( $block[ 'ESCAPE' ] ) ) {
                    $esc = DEFAULT_ESCAPE_METHOD;
                }
                else {
                    $esc = $block[ 'ESCAPE' ];
                }
                switch ( $esc ) {
                    case ESCMD_HTML:
                        $res = htmlspecialchars( $res );
                        break;
                }
            }
            return $res . $block[ 'tail' ];
        }
        // }}} _processVar()

        // {{{ _processLoop()
        /**
         * @param array $block
         * @return string
         */
        function _processLoop( & $block )
        {
            $res = '';
            $values = & $this->_getVar( $block, $block[ 'NAME' ] );
            if ( is_array( $values ) ) {
                reset( $values );
                while ( list( , $value ) = each( $values ) ) {
                    $block[ 'values' ] = & $value;
                    $res .= $block[ 'tail' ] . $this->_processTags( $block[ 'underlying' ] );
                    unset( $block[ 'values' ] );
                }
            }
            return ( $res .= $block[ 'endtail' ] );
        }
        // }}} _processLoop()

        // {{{ _processIf()
        /**
         * @param array $block
         * @return string
         */
        function _processIf( & $block )
        {
            $res = '';
            $s = split( '([^a-zA-Z0-9_]+)', $block[ 'NAME' ] );
            if( sizeof( $s ) > 1 && !strpos( $block[ 'NAME' ], '(' )
                                 && !strpos( $block[ 'NAME' ], '$' ) ) {
                $a = array();
                while( list( $key ) = each( $s ) ) {
                    // пробуем получить переменную по имени
                    $var = $this->_getVar( $block, $s[ $key ] );
                    if( $var === null )
                        unset( $s[ $key ] );
                    elseif( is_string( $var ) )
                        // если это строка, заключаем ее в кавычки
                        $a[ $key ] = '\'' . str_replace( array('\\','\''), array('\\\\','\\\''), $var ) . '\'';
                    else
                        $a[ $key ] = $var;
                }
                $condition = '(' . str_replace( $s, $a, $block[ 'NAME' ] ) . ')==true';
                $values = false;
                eval( '$values=' . $condition . ';' );
                unset( $a );
            }
            else {
                $values = & $this->_getVar( $block, $block[ 'NAME' ] );
            }
            unset( $s );
            if ( $values ) {
                $res .= $block[ 'tail' ] . $this->_processTags( $block[ 'underlying' ] );
            }
            elseif ( isset( $block[ 'else' ] ) || isset( $block[ 'elsetail' ] ) ) {
                $res .= $block[ 'elsetail' ] . $this->_processTags( $block[ 'else' ] );
            }
            return ( $res .= $block[ 'endtail' ] );
        }
        // }}} _processIf()

        // {{{ _processUnless()
        /**
         * @param array $block
         * @return string
         */
        function _processUnless( & $block )
        {
            $res = '';
            $values = & $this->_getVar( $block, $block[ 'NAME' ] );
            if ( !$values ) {
                $res .= $block[ 'tail' ] . $this->_processTags( $block[ 'underlying' ] );
            }
            elseif ( isset( $block[ 'else' ] ) || isset( $block[ 'elsetail' ] ) ) {
                $res .= $block[ 'elsetail' ] . $this->_processTags( $block[ 'else' ] );
            }
            return ( $res .= $block[ 'endtail' ] );
        }
        // }}} _processUnless()

        // {{{ _processInclude()
        /**
         * @param array $block
         * @return string
         */
        function _processInclude( & $block )
        {
            if ( ( strpos($block[ 'NAME' ], '/') !== false ) or ( strpos($block[ 'NAME' ], '\\') !== false ) )
                $dir = '.';
            else
                $dir = $this->_dirname;
            $tmpl = & new CTemplate( $dir, $block[ 'NAME' ] );
            $tmpl->param( $this->_values );
            return ( $tmpl->output() . $block[ 'tail' ] );
        }
        // }}} _processInclude()

    } // class CTemplate

?>
